/*
 * File event notification support.
 *
 * Copyright (C) 2021, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <map>
#include <mutex>

#include <ipc/cap_alloc.h>
#include <ipc/server.h>

#include <pthread.h>
#include <pthread-l4.h>

#include "notification_client.h"
#include "notifier.h"



/* Null notification state. */

static FileNotificationState _null_state;



/* Lock protecting per-task notifier access. */

static std::mutex _lock;

/* Per-task storage for specific waiting operations. */

static SpecificFileNotifier *_notifier = NULL;



/* Return the per-task notifier for file-specific waiting operations. */

SpecificFileNotifier *notifier_get_task_notifier()
{
  std::lock_guard<std::mutex> guard(_lock);

  /* Start any new notifier. */

  if (_notifier == NULL)
  {
    _notifier = new SpecificFileNotifier;
    _notifier->start();
  }

  return _notifier;
}

/* Return a local notifier for general file waiting operations. */

GeneralFileNotifier *notifier_get_local_notifier()
{
  GeneralFileNotifier *notifier = new GeneralFileNotifier;

  notifier->start();
  return notifier;
}



/* Invoke the mainloop in a thread. */

static void *notifier_mainloop(void *data)
{
  FileNotifier *notifier = reinterpret_cast<FileNotifier *>(data);

  notifier->mainloop();
  return 0;
}



/* Virtual destructor required for base class instance reference deletion. */

FileNotifier::~FileNotifier()
{
}

/* Listen for notifications. */

void FileNotifier::mainloop()
{
  ipc_message_t msg;
  l4_umword_t label;

  while (1)
  {
    ipc_message_wait(&msg, &label);

    /* Clear lower label bits. */

    label = label & ~3UL;

    /* Ignore erroneous messages. */

    if (l4_ipc_error(msg.tag, l4_utcb()))
      continue;

    /* Interpret gate labels as file objects. */

    file_t *file = (file_t *) label;

    /* Obtain message details. */

    ipc_message_open(&msg);

    notify_flags_t flags = ipc_message_get_word(&msg, 0);

    /* Reply to notifications. */

    ipc_message_reply(&msg);
    ipc_message_discard(&msg);

    /* Register the notification. */

    _notify(file, flags);
  }

  ipc_message_free(&msg);
}

/* Start listening for notifications. */

long FileNotifier::start()
{
  if (_started)
    return L4_EOK;

  pthread_t thread;
  pthread_attr_t attr;
  long err;

  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

  err = pthread_create(&thread, &attr, notifier_mainloop, this);
  if (err)
    return err;

  _thread = pthread_l4_cap(thread);
  _started = true;

  return L4_EOK;
}



/* Return a notification state object for the given file or a null object if no
   record existed for the file. */

FileNotificationState &FileNotifier::file_state(file_t *file, bool create)
{
  FileNotificationStates::iterator it = _state.find(file);

  if (it == _state.end())
  {
    if (create)
      return _state[file];
    else
      return _null_state;
  }

  return it->second;
}

/* Subscribe to notification events on a file. */

long FileNotifier::subscribe(file_t *file, notify_flags_t flags)
{
  /* Acquire the lock for state lookup. */

  std::unique_lock<std::mutex> state_guard(_state_lock);

  FileNotificationState &state = file_state(file, true);

  /* Create a notification endpoint, if necessary. */

  if (state.is_null())
  {
    long err = ipc_server_new_for_thread(&state.endpoint, file, _thread);

    if (err)
      return err;
  }

  client_Notification notify(file->ref);

  return notify.subscribe(state.endpoint, flags);
}

/* Unsubscribe from notification events on a file. */

long FileNotifier::unsubscribe(file_t *file)
{
  /* Acquire the lock for state lookup. */

  std::unique_lock<std::mutex> state_guard(_state_lock);

  FileNotificationState &state = file_state(file, false);

  if (state.is_null())
    return -L4_EINVAL;

  _state.erase(file);

  /* Remove the lock for updating file state. */

  _file_locks.erase(file);

  return L4_EOK;
}



/* Handle a notification event for a file. Ideally, this would be invoked by the
   generic server dispatch mechanism, with the gate label being interpreted and
   provided as the first parameter. */

void GeneralFileNotifier::_notify(file_t *file, notify_flags_t flags)
{
  /* Enter critical section for the notifier (affecting all files). */

  std::unique_lock<std::mutex> general_guard(_general_lock);

  /* Acquire the lock for state lookup. */

  std::unique_lock<std::mutex> state_guard(_state_lock);

  FileNotificationState &state = file_state(file, false);

  if (state.is_null())
    return;

  /* Acquire the lock for the file state itself. */

  std::unique_lock<std::mutex> file_guard(state.lock);

  /* Record flags and return previous flags. */

  notify_flags_t recorded = state.pending;

  state.pending |= flags;

  /* Add a file queue entry for any files without previous notifications. */

  if (!recorded)
    _affected.push_back(file);

  /* Notify any waiting caller. */

  _general_condition.notify_one();
}

void SpecificFileNotifier::_notify(file_t *file, notify_flags_t flags)
{
  /* Acquire the lock for state lookup. */

  std::unique_lock<std::mutex> state_guard(_state_lock);

  FileNotificationState &state = file_state(file, false);

  if (state.is_null())
    return;

  /* Acquire the lock for the file state itself. */

  std::unique_lock<std::mutex> file_guard(state.lock);

  state.pending |= flags;

  /* Notify any waiting caller. */

  state.condition.notify_one();
}



/* Transfer pending notifications to the given file. This must be called with a
   lock acquired on the file notification state. */

bool FileNotifier::_transfer(FileNotificationState &state, file_t *file)
{
  notify_flags_t recorded = state.pending;

  if (recorded)
  {
    file->notifications = recorded;
    state.pending = 0;
    return true;
  }

  return false;
}



/* Obtain file state and transfer notifications. */

bool GeneralFileNotifier::_retrieve_for_file(file_t *file)
{
  /* Acquire the lock for state lookup. */

  std::unique_lock<std::mutex> state_guard(_state_lock);

  FileNotificationState &state = file_state(file, false);

  if (state.is_null())
    return false;

  /* Acquire the lock for the file state itself, then release the state lock. */

  std::unique_lock<std::mutex> file_guard(state.lock);

  state_guard.unlock();

  /* Call generic method to transfer notifications, if possible. */

  return _transfer(state, file);
}

/* Obtain queued files until one is found that still has events recorded for it.
   This must be called with the notifier's general lock acquired. */

bool GeneralFileNotifier::_retrieve(file_t **file)
{
  while (!_affected.empty())
  {
    *file = _affected.front();
    _affected.pop_front();

    if (_retrieve_for_file(*file))
      return true;
  }

  return false;
}



/* Wait for notification events on files. */

long GeneralFileNotifier::wait(file_t **file)
{
  std::unique_lock<std::mutex> general_guard(_general_lock);

  while (1)
  {
    /* With pending notifications, update the first file and exit. */

    if (_retrieve(file))
      break;

    /* Otherwise, wait for notifications. */

    _general_condition.wait(general_guard);
  }

  return L4_EOK;
}

/* Wait for notifications from a single file. */

long SpecificFileNotifier::wait_file(file_t *file)
{
  /* Acquire the lock for reading file state. */

  std::unique_lock<std::mutex> state_guard(_state_lock);

  FileNotificationState &state = file_state(file, false);

  if (state.is_null())
    return -L4_EINVAL;

  /* Acquire the lock for the file state itself, then release the state lock. */

  std::unique_lock<std::mutex> file_guard(state.lock);

  state_guard.unlock();

  while (1)
  {
    /* With pending notifications, update the file and exit. */

    if (_transfer(state, file))
      break;

    /* Otherwise, wait for notifications. */

    state.condition.wait(file_guard);
  }

  return L4_EOK;
}

// vim: tabstop=2 expandtab shiftwidth=2
